# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import Optional

from sqllineage.core.models import Table
from sqllineage.runner import LineageRunner

from vdk.api.lineage.sql.data import LineageTable, LineageData

log = logging.getLogger(__name__)


def get_table_lineage_from_query(
        query: str, schema: Optional[str], catalog: Optional[str]
) -> LineageData:
    """
    This method parses the sql query. If and only if it is a rename table query,
    the method returns the names of the source and destination table.
    :param query: The SQL query potentially containing a RENAME TABLE operation
    :param schema: The schema which is queried
    :param catalog: The catalog which is queried
    :return: A tuple with (table_from, table_to) if it is a RENAME TABLE query, None otherwise.
    """
    runner = LineageRunner(query)

    if (len(runner.statements_parsed) == 0):
        #log.debug("No statement passed")
        return None

    if len(runner.statements_parsed) > 1:
        raise RuntimeError("Query with more than one statement is passed. "
                           "Make sure that multiple query statements (separated) are not passed to this method."
                           f"Query passed is '{query}' ")

    input_tables = [_lineage_table_from_name(source_table, schema, catalog) for source_table in runner.source_tables]
    output_table = None
    if len(runner.target_tables) == 1:
        output_table = _lineage_table_from_name(runner.target_tables[0], schema, catalog)
    elif len(runner.target_tables) > 1:
        raise RuntimeError("Query with more than one target table should not be possible. "
                           "Make sure that multiple queries (separated) are not passed accidentally."
                           f"Query is '{query}' ")

    return LineageData(
        query=query,
        query_type=runner.statements_parsed[0].get_type(),
        query_status="",
        input_tables=input_tables,
        output_table=output_table,
    )


def _lineage_table_from_name(
        table: Table, schema: str, catalog: str
) -> LineageTable:
    tokens = []
    if table.schema:
        tokens = table.schema.raw_name.split(".")

    if len(tokens) == 0:
        return LineageTable(catalog, schema, table.raw_name)
    elif len(tokens) == 1:
        return LineageTable(catalog, tokens[0], table.raw_name)
    elif len(tokens) == 2:
        return LineageTable(tokens[0], tokens[1], table.raw_name)
    else:
        return None



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass
from typing import List


@dataclass(frozen=True)
class LineageTable:
    """
    Defines the LineageTable object contract

    Attributes
    catalog: str
        A string holding the catalog where the table lives
    schema: str
        A string holding the schema where the table lives
    table: str
        A string holding the table name
    """

    catalog: str
    schema: str
    table: str

    def __str__(self):
        s = ""
        if self.catalog:
            s = f"{self.catalog}."
        if self.schema:
            s = f"{s}{self.schema}."
        return f"{s}{self.table}"


@dataclass(frozen=True)
class LineageData:
    """
    Defines the LineageData contract

    Attributes
    query: str
        The original query
    query_type: str
        Type of operation (see below for supported values)
    query_status: str
        'OK' or 'EXCEPTION'
    input_tables: List[LineageTable]
        An array of LineageTable objects (see LineageTable)
    output_table: LineageTable
        A LineageTable object (see LineageTable)

    Supported query_type values are:
        - insert
        - select
        - insert_select
        - rename_table
    """

    query: str
    query_type: str
    query_status: str
    input_tables: List[LineageTable]
    output_table: LineageTable


from abc import ABC

from vdk.api.lineage.sql.data import LineageData


class LineageLogger(ABC):
    """
    This interface describes what behaviour a lineage logger must possess to interact with the lineage logging
    functionality afforded by different plugins.
    """

    def send(self, lineage_data: LineageData) -> None:
        """
        This method sends the collected lineage data to some lineage data processing application.

        :param lineage_data: The collected lineage data.
        """
        pass



import uuid
from datetime import datetime

import attr
from openlineage.client.client import OpenLineageClient, OpenLineageClientOptions
from openlineage.client.facet import ParentRunFacet, SqlJobFacet, BaseFacet, SCHEMA_URI
from openlineage.client.run import RunEvent, RunState, Run, Job, Dataset
from typing import Optional

from vdk.plugin.lineage import sql_lineage_parser

PRODUCER = "https://github.com/vmware/versatile-data-kit/tree/main/projects/vdk-plugins/vdk-lineage"



@attr.s
class VdkJobFacet(BaseFacet):
    op_id: str = attr.ib()
    execution_id: str = attr.ib()
    source_version: str = attr.ib()
    team: str = attr.ib()

    @staticmethod
    def _get_schema() -> str:
        return SCHEMA_URI + "#/definitions/VdkJobFacet"


def setup_client(url: str, api_key: str = None) -> Optional[OpenLineageClient]:
    if not url:
        return None
    return OpenLineageClient(
        url=url,
        options=OpenLineageClientOptions(
            api_key=api_key
        )
    )


def run_event(
        state: RunState,
        job_name: str,
        job_namespace: str,
        run_id: str = str(uuid.uuid4()),
        parent: Optional[ParentRunFacet] = None,
        run_details: Optional[VdkJobFacet] = None
) -> RunEvent:
    return RunEvent(
        eventType=state,
        eventTime=datetime.now().isoformat(),
        run=Run(
            runId=run_id,
            facets={
                "parent": parent
            } if parent else {}
        ),
        job=Job(
            namespace=parent if parent else job_namespace,
            name=job_name,
            facets={"details": run_details} if run_details else {}
        ),
        producer=PRODUCER
    )


def sql_event(state: RunState,
              sql: str,
              job_namespace: str,
              run_id: str,
              parent: Optional[ParentRunFacet] = None):
    lineage_data = sql_lineage_parser.get_table_lineage_from_query(sql, None, None)
    inputs = [Dataset(job_namespace, str(t)) for t in lineage_data.input_tables]
    outputs = [Dataset(job_namespace, str(lineage_data.output_table))] if lineage_data.output_table else []

    return RunEvent(
        eventType=state,
        eventTime=datetime.now().isoformat(),
        run=Run(
            runId=run_id,
            facets={
                "parent": parent
            } if parent else {}
        ),
        job=Job(
            namespace=job_namespace,
            name=f"{parent.job['name']}",
            facets={
                'sql': SqlJobFacet(sql)
            }
        ),
        inputs=inputs,
        outputs=outputs,
        producer=PRODUCER
    )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import pathlib

import setuptools

"""
Builds a package with the help of setuptools in order for this package to be imported in other projects

"plugin-package-template" is the name of the package which contains the plugin/s.

"vdk-lineage" is the name of the plugin contained within this package. Note that you can include
more than one plugin in a single package. Also note that the contained plugin may have the same name as the package.

"plugin_template" is the Python file which contains the plugin hooks for the corresponding plugin.
"""

__version__ = "0.1.3"

setuptools.setup(
    name="vdk-lineage",
    version=__version__,
    url="https://github.com/vmware/versatile-data-kit",
    description="Plugin template project used to quick start development of a new Versatile Data Kit SDK plugin.",
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    install_requires=["vdk-core", "sqlparse", "sqllineage",
                      "openlineage-integration-common", "openlineage-python"],
    package_dir={"": "src"},
    packages=setuptools.find_namespace_packages(where="src"),
    entry_points={
        "vdk.plugin.run": [
            "vdk-lineage = vdk.plugin.lineage.plugin_lineage"
        ]
    },
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



import logging
import uuid
from typing import List, Optional

import pluggy
from openlineage.client import OpenLineageClient
from openlineage.client.facet import ParentRunFacet
from openlineage.client.run import RunState
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.config.job_config import JobConfigKeys
from vdk.internal.builtin_plugins.connection.decoration_cursor import DecorationCursor
from vdk.internal.builtin_plugins.run.execution_results import ExecutionResult
from vdk.internal.builtin_plugins.run.execution_state import ExecutionStateStoreKeys
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.core.context import CoreContext
from vdk.internal.core.statestore import CommonStoreKeys
from vdk.internal.plugin.plugin import PluginRegistry

from vdk.plugin.lineage import openlineage_config
from vdk.plugin.lineage.openlineage_config import OpenLineageConfiguration
from vdk.plugin.lineage.openlineage_utils import setup_client, run_event, VdkJobFacet, sql_event

log = logging.getLogger(__name__)


class OpenLineagePlugin:

    def __init__(self):
        self.__client: Optional[OpenLineageClient] = None
        self.__op_id = None
        self.__execution_id = None
        self.__job_version = None
        self.__team = None
        self.__job_name = None
        self.__namespace = "versatile_data_kit"
        self.__job_run_id = str(uuid.uuid4())

    @hookimpl
    def vdk_configure(self, config_builder: ConfigurationBuilder) -> None:
        """
        Here we define what configuration settings are needed for openlineage.
        """
        openlineage_config.add_definitions(config_builder)

    @hookimpl
    def vdk_initialize(self, context: CoreContext):
        config = OpenLineageConfiguration(context.configuration)
        if config.url():
            log.debug(f"Open lineage client is enabled at url: {config.url()}")
            self.__client = setup_client(config.url(), config.api_key())
        else:
            log.debug("No OpenLineage url set. Collecting openlineage data is disabled.")

    @hookimpl
    def initialize_job(self, context: JobContext) -> None:
        if self.__client:
            self.__execution_id = context.core_context.state.get(CommonStoreKeys.EXECUTION_ID)
            self.__op_id = context.core_context.state.get(CommonStoreKeys.OP_ID)
            self.__job_version = context.core_context.state.get(ExecutionStateStoreKeys.JOB_GIT_HASH)
            self.__team = context.core_context.configuration.get_value(JobConfigKeys.TEAM)
            self.__job_name = context.job_input.get_name()

            self.__emit_run_event(RunState.START)

    @hookimpl(hookwrapper=True)
    def run_job(self, context: JobContext) -> Optional[ExecutionResult]:
        out: pluggy.callers._Result
        out = yield
        if self.__client:
            result: ExecutionResult = out.get_result()
            self.__execution_id = context.core_context.state.get(CommonStoreKeys.EXECUTION_ID)
            self.__op_id = context.core_context.state.get(CommonStoreKeys.OP_ID)
            self.__emit_run_event(RunState.COMPLETE if result.is_success() else RunState.FAIL)
        return None

    @hookimpl(trylast=True)
    def db_connection_decorate_operation(
            self, decoration_cursor: DecorationCursor
    ):
        if self.__client:
            # TODO: track properly the run state.
            # For now it's always complete since we have not easy way to know start/end.
            query = decoration_cursor.get_managed_operation().get_operation()
            try:
                #  TODO: abstract away so there could be multiple lineage loggers (nit just openlineage)
                self.__client.emit(sql_event(RunState.COMPLETE, query, self.__namespace,
                                             self.__job_run_id,
                                             ParentRunFacet.create(self.__job_run_id, self.__namespace, self.__job_name)))
            except Exception as e:
                log.exception(f"Failed to send lineage data for query {query}. Will continue without it. Errors was {e}")
                raise e

    def __emit_run_event(self, state: RunState):
        try:
            run_details = VdkJobFacet(self.__op_id, self.__execution_id, self.__job_version, self.__team)
            self.__client.emit(run_event(state, self.__job_name, self.__namespace,
                                         self.__job_run_id, None, run_details))
        except Exception as e:
            log.exception(f"Failed to send lineage data for the job. Will continue without it. Errors was {e}.")


@hookimpl
def vdk_start(plugin_registry: PluginRegistry, command_line_args: List) -> None:
    """
    Load all default (builtin) vdk plugins
    """
    log.debug("Load default (builtin) vdk plugins.")
    plugin_registry.load_plugin_with_hooks_impl(OpenLineagePlugin())



from vdk.internal.core.config import Configuration, ConfigurationBuilder

OPENLINEAGE_API_KEY = "OPENLINEAGE_API_KEY"
OPENLINEAGE_URL = "OPENLINEAGE_URL"


class OpenLineageConfiguration:

    def __init__(self, config: Configuration):
        self.__config = config

    def url(self):
        return self.__config.get_value(OPENLINEAGE_URL)

    def api_key(self):
        return self.__config.get_value(OPENLINEAGE_API_KEY)


def add_definitions(config_builder: ConfigurationBuilder):
    config_builder.add(
        key=OPENLINEAGE_URL,
        default_value=None,
        description="The URL of the service which will consume OpenLineage events.",
    )
    config_builder.add(
        key=OPENLINEAGE_API_KEY,
        default_value=None,
        description="The `Bearer` authentication key used if required by consumer service of OpenLineage events.",
    )
